/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 2/25/2001                                                  */
/*                                                                           */
/* File: "histogram.c"                                                       */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef INCLUDE_JAVA
#include "GatorCameraClient.h"
#endif

#include "acticast_include.h"

/****************************** Subroutines **********************************/

int floor_histograms(double **hist, int n_bins, int n_objects, double floor)
{
  int i, j;

  for(i = 0; i < n_bins; i++) 
    for(j = 0; j < n_objects; j++) {
      if(is_zero2(hist[i][j], floor))
	hist[i][j] = floor;
    }

  normalize_histograms(hist, n_bins, n_objects);

  return(1);
}

/*****************************************************************************/

int normalize_histograms(double **hist, int n_bins, int n_objects)
{
  int i, j;
  double sum;
  
  for(i = 0; i < n_bins; i++) {
    sum = 0.;

    for(j = 0; j < n_objects; j++) {
      sum += hist[i][j];
    }

    for(j = 0; j < n_objects; j++) {
      hist[i][j] /= sum;
    }
  }
    
  return(1);
}

/*****************************************************************************/

int vector_quantize(double r, double g, double b, double **vq, int n)
{
  int i, d, dindex, dmin;

  dmin = 195075; // (255*255*3)

  for(i = 0; i < n; i++) {
    d = (vq[i][0] - r)*(vq[i][0] - r) +
      (vq[i][1] - g)*(vq[i][1] - g) +
      (vq[i][2] - b)*(vq[i][2] - b);

    if(d < dmin) {
      dindex = i;
      dmin = d;
    }
  }

  return(dindex);
}

/*****************************************************************************/

int int_vector_quantize(int r, int g, int b, double **vq, int n)
{
  double rd, gd, bd;

  rd = (double) r;
  gd = (double) g;
  bd = (double) b;

  return(vector_quantize(rd, gd, bd, vq, n));
}

/*****************************************************************************/

int compute_histogram_class(BLOB2 *b, double **vq, int n_vq, 
			    double **hist, int n_hist)
{
  int i, index;

  for(i = 0; i < b->n; i++) {
    index = vector_quantize(b->rgb_t[RED][i], 
			    b->rgb_t[GREEN][i],
			    b->rgb_t[BLUE][i], vq, n_vq);

    b->type[i] = ind_max_vec(&hist[index][0], n_hist);
  }

  return(1);
}

/*****************************************************************************/

int compute_histogram_prob_vec(double *p_vec,
			       BLOB2 *b, double **vq, int n_vq, 
			       double **hist, int n_hist)
{
  int i;

  for(i = 0; i < n_hist; i++)
    p_vec[i] = 0.;

  compute_histogram_class(b, vq, n_vq, hist, n_hist);

  for(i = 0; i < b->n; i++) {
    p_vec[b->type[i]] += 1.;
  }

  for(i = 0; i < n_hist; i++)
    p_vec[i] /= ((double) b->n);

  //fprintf(stderr, "p_vec = %lf %lf\n", p_vec[0], p_vec[1]);

  return(1);
}

/*****************************************************************************/

// experimental 

int compute_histogram_prob_vec2(double *p_vec,
				BLOB2 *b, double **vq, int n_vq, 
				double **hist, int n_hist)
{
  int i, j, index;
  double min_val, sum;

  for(j = 0; j < n_hist; j++)
    p_vec[j] = 0.;

  for(i = 0; i < b->n; i++) {
    index = vector_quantize(b->rgb_t[RED][i], 
			    b->rgb_t[GREEN][i],
			    b->rgb_t[BLUE][i], vq, n_vq);

    b->type[i] = ind_max_vec(&hist[index][0], n_hist);

    sum = 0.;
    for(j = 0; j < n_hist; j++)
      sum += hist[index][j];

    for(j = 0; j < n_hist; j++)
      p_vec[j] += hist[index][j]/sum;
  }

  sum = 0.;
  for(j = 0; j < n_hist; j++) {
    sum += p_vec[j];
  }
  

  for(j = 0; j < n_hist; j++)
    p_vec[j] /= sum;

  return(1);
}

/*****************************************************************************/

int generate_multinomial_hypotheses(int **hyp, int n, int m)
{
  int i, size, v[N_OBJECTS_MAX];

  for(i = 0; i < m; i++)
    v[i] = i;
    
  size = generate_hypotheses(hyp, &v[0], n, m, 0);
  // display_int_matrix(stdout, hyp, size, m);
  return(size);
}

/*****************************************************************************/

int generate_hypotheses(int **hyp, int *v, int n, int m, int count)
{
  int i, tst;

  for(i = 0; i < m; i++)
    hyp[count][i] = v[i];
  count++;

  if(v[m-1] >= (n-1)) {
    tst = -1;
    for(i = (m-2); i >= 0; i--) {
      if(v[i]+1 <= (n-m+i)) {
	tst = i;
	break;
      }
    }

    if(tst >= 0) {
      v[tst]++;
      
      for(i = tst+1; i < m; i++) {
	v[i]=v[i-1]+1;
      }
      generate_hypotheses(hyp, v, n, m, count);
    } else {
      return(count);
    }
  } else {
    v[m-1]++;
    generate_hypotheses(hyp, v, n, m, count);
  }
}

/*****************************************************************************/

double test_hypothesis(int *hyp, int n_hyp, double *p)
{
  int i;
  double log_sum, expon;

  log_sum = log((double) n_hyp);
  expon = 1./((double) n_hyp);

  for(i = 0; i < n_hyp; i++)
    log_sum += expon*log(p[hyp[i]]);

  return(log_sum);
}

/*****************************************************************************/

int which_hypotheses_is_most_probable(int *ans, int **hyp, BLOB2 *b, 
				      double **vq, int n_vq, 
				      double **hist, int n_hist)
{
  int i, j, k, l;
  double p[N_OBJECTS_MAX];
  double n_hyp, n_objs;
  double tst, max = -1000000.;

  if(b->type[0] < 0) 
    compute_histogram_class(b, vq, n_vq, hist, n_hist);

  for(i = 0; i < n_hist; i++)
    p[i] = 0.;

  for(j = 0; j < b->n; j++) {
    p[ind_max_vec(&hist[b->type[j]][0],3)]+=1.;
  }

  for(i = 0; i < n_hist; i++)
    p[i] /= ((double) b->n);

  /*
    for(i = 0; i < n_hist; i++) {
    p[i] = 0.;
    for(j = 0; j < b->n; j++) {
    p[i] += hist[b->type[j]][i];
    }
    p[i] /= ((double) b->n);
    }
  */

  fprintf(stdout, "p: %.7f %.7f %.7f\n", p[0], p[1], p[2]);

  for(i = 0; i < n_hist; i++) {
    n_hyp = generate_multinomial_hypotheses(hyp, 3, i + 1);
    for(j = 0; j < n_hyp; j++) {
      tst = test_hypothesis(&hyp[j][0], i + 1, &p[0]);
      if(tst > max) {
	for(k = 0; k < i+1; k++)
	  ans[k] = hyp[j][k];
	max = tst;
	n_objs = i + 1;
      }
#ifdef LEVEL_THREE_DIAGNOSTIC
      fprintf(stdout, "hypothesis = {");
      for(k = 0;  k < i; k++)
	fprintf(stdout, "%d, ", hyp[j][k]);
      fprintf(stdout, "%d}:", hyp[j][i]);
      fprintf(stdout, "%.15f\n", tst);
#endif

    }
  }

  return(n_objs);
}

/*****************************************************************************/

/* assumes all blobs are found */

/* specific to 1 2 or three objects -- working on generalization */

int generate_hypothesis_struct(HYPOTHESES *all, int n_objs, int n_blobs)
{
  int i, j;

  if(n_blobs > n_objs) // Terrible hack to deal with split objects
    n_blobs = n_objs;

  switch(n_objs) 
    {
    case 1: 
      {
	all->n_hyp = 1;
	all->hyp[0].n = 1;
	all->hyp[0].leaf[0].n_objs = 1;
	all->hyp[0].leaf[0].objs[0] = 0;
	break;
      }
    case 2:
      {
	switch(n_blobs)
	  {
	  case 1: // Modified to allow for missing cars -- generalize
	    {
	      all->n_hyp = 3;

	      all->hyp[0].n = 1;
	      all->hyp[0].leaf[0].n_objs = 1;
	      all->hyp[0].leaf[0].objs[0] = 0;

	      all->hyp[1].n = 1;
	      all->hyp[1].leaf[0].n_objs = 1;
	      all->hyp[1].leaf[0].objs[0] = 1;

	      all->hyp[2].n = 1;
	      all->hyp[2].leaf[0].n_objs = 2;
	      all->hyp[2].leaf[0].objs[0] = 0;
	      all->hyp[2].leaf[0].objs[1] = 1;

	      /* don't allow meged hypothesis */

	      //all->n_hyp = 2;

	      /* allow only merged hypothesis if one car is visible */

	      // all->n_hyp = 1;
	      // all->hyp[0].n = 1;
	      // all->hyp[0].leaf[0].n_objs = 2;
	      // all->hyp[0].leaf[0].objs[0] = 0;
	      // all->hyp[0].leaf[0].objs[1] = 1;
	      break;
      	    }
	  case 2: 
	    {
	      all->n_hyp = 2;
	      all->hyp[0].n = 2;
	      all->hyp[0].leaf[0].n_objs = 1;
	      all->hyp[0].leaf[0].objs[0] = 0;
	      all->hyp[0].leaf[1].n_objs = 1;
	      all->hyp[0].leaf[1].objs[0] = 1;
	      
	      all->hyp[1].n = 2;
	      all->hyp[1].leaf[0].n_objs = 1;
	      all->hyp[1].leaf[0].objs[0] = 1;
	      all->hyp[1].leaf[1].n_objs = 1;
	      all->hyp[1].leaf[1].objs[0] = 0;
	      break;
	    }
	  default:
	    {
	      fprintf(stderr, "n_objss == 2 && n_blobs != 1 or 2. Exiting.\n");
	      exit(-1);
	    }
	  }
	break;
      }
    case 3:
      {
	switch(n_blobs)
	  {
	  case 1:
	    {
	      all->n_hyp = 1;

	      all->hyp[0].n = 1;
	      all->hyp[0].leaf[0].n_objs = 3;
	      all->hyp[0].leaf[0].objs[0] = 0;
	      all->hyp[0].leaf[0].objs[1] = 1;
	      all->hyp[0].leaf[0].objs[2] = 2;
	      break;
	    }
	  case 2: 
	    {
	      all->n_hyp = 6;
	      
	      all->hyp[0].n = 2;
	      all->hyp[0].leaf[0].n_objs = 1;
	      all->hyp[0].leaf[0].objs[0] = 0;
	      all->hyp[0].leaf[1].n_objs = 2;
	      all->hyp[0].leaf[1].objs[0] = 1;
	      all->hyp[0].leaf[1].objs[1] = 2;

	      all->hyp[1].n = 2;
	      all->hyp[1].leaf[0].n_objs = 1;
	      all->hyp[1].leaf[0].objs[0] = 1;
	      all->hyp[1].leaf[1].n_objs = 2;
	      all->hyp[1].leaf[1].objs[0] = 0;
	      all->hyp[1].leaf[1].objs[1] = 2;

	      all->hyp[2].n = 2;
	      all->hyp[2].leaf[0].n_objs = 1;
	      all->hyp[2].leaf[0].objs[0] = 2;
	      all->hyp[2].leaf[1].n_objs = 2;
	      all->hyp[2].leaf[1].objs[0] = 0;
	      all->hyp[2].leaf[1].objs[1] = 1;

	      all->hyp[3].n = 2;
	      all->hyp[3].leaf[1].n_objs = 1;
	      all->hyp[3].leaf[1].objs[0] = 0;
	      all->hyp[3].leaf[0].n_objs = 2;
	      all->hyp[3].leaf[0].objs[0] = 1;
	      all->hyp[3].leaf[0].objs[1] = 2;

	      all->hyp[4].n = 2;
	      all->hyp[4].leaf[1].n_objs = 1;
	      all->hyp[4].leaf[1].objs[0] = 1;
	      all->hyp[4].leaf[0].n_objs = 2;
	      all->hyp[4].leaf[0].objs[0] = 0;
	      all->hyp[4].leaf[0].objs[1] = 2;

	      all->hyp[5].n = 2;
	      all->hyp[5].leaf[1].n_objs = 1;
	      all->hyp[5].leaf[1].objs[0] = 2;
	      all->hyp[5].leaf[0].n_objs = 2;
	      all->hyp[5].leaf[0].objs[0] = 0;
	      all->hyp[5].leaf[0].objs[1] = 1;
	      break;
	    }
	  case 3:
	    {
	      all->n_hyp = 6;
	      
	      all->hyp[0].n = 3;
	      all->hyp[0].leaf[0].n_objs = 1;
	      all->hyp[0].leaf[0].objs[0] = 0;
	      all->hyp[0].leaf[1].n_objs = 1;
	      all->hyp[0].leaf[1].objs[0] = 1;
	      all->hyp[0].leaf[2].n_objs = 1;
	      all->hyp[0].leaf[2].objs[0] = 2;

	      all->hyp[1].n = 3;
	      all->hyp[1].leaf[0].n_objs = 1;
	      all->hyp[1].leaf[0].objs[0] = 0;
	      all->hyp[1].leaf[1].n_objs = 1;
	      all->hyp[1].leaf[1].objs[0] = 2;
	      all->hyp[1].leaf[2].n_objs = 1;
	      all->hyp[1].leaf[2].objs[0] = 1;

	      all->hyp[2].n = 3;
	      all->hyp[2].leaf[0].n_objs = 1;
	      all->hyp[2].leaf[0].objs[0] = 1;
	      all->hyp[2].leaf[1].n_objs = 1;
	      all->hyp[2].leaf[1].objs[0] = 0;
	      all->hyp[2].leaf[2].n_objs = 1;
	      all->hyp[2].leaf[2].objs[0] = 2;

	      all->hyp[3].n = 3;
	      all->hyp[3].leaf[0].n_objs = 1;
	      all->hyp[3].leaf[0].objs[0] = 2;
	      all->hyp[3].leaf[1].n_objs = 1;
	      all->hyp[3].leaf[1].objs[0] = 0;
	      all->hyp[3].leaf[2].n_objs = 1;
	      all->hyp[3].leaf[2].objs[0] = 1;

	      all->hyp[4].n = 3;
	      all->hyp[4].leaf[0].n_objs = 1;
	      all->hyp[4].leaf[0].objs[0] = 1;
	      all->hyp[4].leaf[1].n_objs = 1;
	      all->hyp[4].leaf[1].objs[0] = 2;
	      all->hyp[4].leaf[2].n_objs = 1;
	      all->hyp[4].leaf[2].objs[0] = 0;

	      all->hyp[5].n = 3;
	      all->hyp[5].leaf[0].n_objs = 1;
	      all->hyp[5].leaf[0].objs[0] = 2;
	      all->hyp[5].leaf[1].n_objs = 1;
	      all->hyp[5].leaf[1].objs[0] = 1;
	      all->hyp[5].leaf[2].n_objs = 1;
	      all->hyp[5].leaf[2].objs[0] = 0;
	      break;
	    }
	  default:
	    {
	      fprintf(stderr, 
		      "n_objs == 3 && n_blobs != 1, 2 or 3. Exiting.\n");
	      exit(-1);
	    }
	  }
	break;
      }
    default:
      {
	fprintf(stderr, "n_objs > 3. Exiting.\n");
	exit(-1);
      }
    }

  return(1);
}

/*****************************************************************************/

double test_one_hypothesis_struct(ONE_HYPOTHESIS *h, double **prob_table)
{
  int i;
  double sum = 0.;

  for(i = 0; i < h->n; i++)
    sum += test_one_hypothesis_leaf(&h->leaf[i], i, prob_table);
      
  return(sum);
}

/*****************************************************************************/

double test_one_hypothesis_leaf(HYPOTHESIS_LEAF *l, int n, double **prob_table)
{
  int i;
  double sum = 0.;

  for(i = 0; i < l->n_objs; i++)
    sum += fabs(prob_table[n][l->objs[i]] - 1./((double) l->n_objs));

  return(sum);
}

/*****************************************************************************/

int compute_most_consistent_hypothesis(HYPOTHESES *all, int n, int m, 
				       double **prob_table)
{
  int i, min_ind;
  double tst, min_dist = 1000000.;

  generate_hypothesis_struct(all, n, m);

  //fprintf(stderr, "-----------------------\n");
  for(i = 0; i < all->n_hyp; i++) {
    tst = test_one_hypothesis_struct(&all->hyp[i], prob_table);
    if(tst < min_dist) {
      min_ind = i;
      min_dist = tst;
    }
    //fprintf(stderr, "hypothesis %d: %lf\n", i, tst);
  }

  return(min_ind);
}

/*****************************************************************************/

#ifndef N_CYCLES_POSITION_ITERATION
#define N_CYCLES_POSITION_ITERATION 5
#endif

int resolve_hypothesis(ONE_HYPOTHESIS *h, BLOB2 *blobs)
{
  int i, j;

  for(i = 0; i < h->n; i++) {
    if(h->leaf[i].n_objs == 1) {
      for(j = 0; j < blobs[i].n; j++) {
	blobs[i].type[j] = h->leaf[i].objs[0];
      }
      blobs[i].tags[0] = h->leaf[i].objs[0];
      blobs[i].n_tags  = 1;
    } else {
      for(j = 0; j < N_CYCLES_POSITION_ITERATION; j++) {
	compute_position_class(&blobs[i], 
			       &h->leaf[i].objs[0], h->leaf[i].n_objs);
      }
    }
  }

  return(1);
}

/*****************************************************************************/
