/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 5/17/2003                                                  */
/*                                                                           */
/* File: "hough.c"                                                           */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "acticast_include.h"
#include "hough.h"

/*****************************************************************************/

int define_gaussian_filter(FILTER *f, int size)
{
  int i, j;
  double sigma, scale;

  f->size = size;
  
  sigma = ((double) size)/4.;
  scale = 1./(2.*M_PI*sigma*sigma);

  for(i = -size; i <= size; i++)
    for(j = -size; j <= size; j++)
      f->m[i+size][j+size] = 
	scale*exp(-.5*((double) (i*i + j*j))/(sigma*sigma));
    
  return(1);
}

/*****************************************************************************/

int sobel_operator(PPM *imin, PPM *imout)
{
  int i, j, k;
  double a11, a12, a13, a21, a23, a31, a32, a33;
  double gx, gy, grad;

  limit_dimensions(imin);
  copy_dimensions(imin, imout);
  
  for(i = 0; i < imin->ydim; i++) {
    for(j = 0; j < imin->xdim; j++) {
      if(i == 0 || j == 0 || i == imin->ydim-1 || j == imin->xdim-1) {
	imout->p[j][i][0] = (char) 0;
	imout->p[j][i][1] = (char) 0;
	imout->p[j][i][2] = (char) 0;
      } else {

	for(k = 0; k < 3; k++) {
	  a11 = ((double) convert_char(imin->p[j-1][i-1][k]));
	  a12 = ((double) convert_char(imin->p[j][i-1][k]));
	  a13 = ((double) convert_char(imin->p[j+1][i-1][k]));
	  a21 = ((double) convert_char(imin->p[j-1][i][k]));
	  a23 = ((double) convert_char(imin->p[j+1][i][k]));
	  a31 = ((double) convert_char(imin->p[j-1][i+1][k]));
	  a32 = ((double) convert_char(imin->p[j][i+1][k]));
	  a33 = ((double) convert_char(imin->p[j+1][i+1][k]));
	  
	  gx = -a11 + a13 - 2.*a21 + 2.*a23 - a31 + a33;
	  gy =  a11 + 2.*a12 + a13 - a31 - 2.*a32 - a33;
	  
	  grad = sqrt(.125*(gx*gx + gy*gy));

	  imout->p[j][i][k] = (char) (grad + .5);
	}
      }
    }
  }

  return(1);
}

/*****************************************************************************/

int hough_transform(PPM *gim, PPM *thim, HOUGH_ACCUMULATOR *hough)
{
  int n, m, i, j, k, test, rho_index;
  double xdim, ydim, *ct, *st, theta, rho, grad;

  xdim = ((double) gim->xdim);
  ydim = ((double) gim->ydim);

  n = hough->theta_partitions;
  m = hough->rho_partitions;

  hough->theta_min = 0.;
  hough->theta_max = M_PI;

  hough->rho_min = -sqrt(xdim*xdim + ydim*ydim);
  hough->rho_max = -(hough->rho_min);

  hough->theta_step = 
    (hough->theta_max - hough->theta_min)/((double) n);
  hough->rho_step = 
    (hough->rho_max - hough->rho_min)/((double) m);

  ct = allocate_vector(ct, n);
  st = allocate_vector(st, n);
  hough->accumulator = allocate_matrix(hough->accumulator, n, m);
  zero_matrix(hough->accumulator, n, m);

  theta = hough->theta_step/2.;
  for(k = 0; k < n; k++) {
    ct[k] = cos(theta);
    st[k] = sin(theta);
    theta += hough->theta_step;
  }

  for(i = 0; i < gim->ydim; i++) {
    for(j = 0; j < gim->xdim; j++) {
      test = convert_char(thim->p[j][i][0]);
      if(test > 0) {
	grad = (double) 
	  (convert_char(gim->p[j][i][0]) +
	   convert_char(gim->p[j][i][1]) +
	   convert_char(gim->p[j][i][2]));
	for(k = 0; k < n; k++) {
	  rho = ((double) j)*ct[k] + ((double) i)*st[k];
	  rho_index = (int)
	    ((rho - hough->rho_min)/(hough->rho_max - hough->rho_min)*
	     ((double) m));
	  hough->accumulator[k][rho_index] += grad;
	}
      }
    }
  }

  free_vector(ct);
  free_vector(st);

  return(1);
}

/*****************************************************************************/

int find_line(HOUGH_ACCUMULATOR *hough, double *slope, double *intercept)
{
  int i, j, imax, jmax;
  double theta, rho, fmax = -1.;


  for(i = 0; i < hough->theta_partitions; i++) {
    for(j = 0; j < hough->rho_partitions; j++) {
      if(hough->accumulator[i][j] > fmax) {
	fmax = hough->accumulator[i][j];
	imax = i;
	jmax = j;
      }
    }
  }

  theta = ((double) imax)*hough->theta_step + hough->theta_step/2.;
  rho = ((double) jmax)*hough->rho_step + hough->rho_min + hough->rho_step/2.;

  *slope = -cos(theta)/sin(theta);
  *intercept = rho/sin(theta);

  fprintf(stderr, "theta = %lf, rho = %lf, m = %lf, b = %lf : %lf\n", theta, 
	  rho, *slope, *intercept, fmax);

  hough->accumulator[imax][jmax] = 0.;

  return(1);
}

/*****************************************************************************/

int plotline(PPM *im, double m, double b, int rv, int gv, int bv)
{
  int i,j;
  double x, y, xres, yres;
  POINT p;

  xres = (double) X_DIM;
  yres = (double) Y_DIM;
  
  if(fabs(m) < 0.75) {
    for(i=0;i<xres;i++) {
      y = i*m+b;
      if(y<yres && y>=0) {
	p.x = (int)i;
	p.y = (int) (y+.5);
	draw_point(im, &p, rv, gv, bv);
	p.x = (int)i;
	p.y = (int)(y+1.5);
	draw_point(im, &p, rv, gv, bv);
      }
    }
  } else {
    for(j=0;j<yres;j++) {
      x=(j-b)/m;
      if(x<xres && x>=0) {
	p.x = (int)(x+.5);
	p.y = (int)j;
	draw_point(im, &p, rv, gv, bv);
	p.x = (int)(x+1.5);
	p.y = (int)j;
	draw_point(im, &p, rv, gv, bv);
      }
    }
  }
}

/*****************************************************************************/

int draw_hough_accumulator(PPM *image, HOUGH_ACCUMULATOR *hough)
{
  matrix_density_plot(image, hough->accumulator, 
		      hough->theta_partitions, hough->rho_partitions,
		      1, COLOR_BLACK);
      
  return(1);
}

/*****************************************************************************/

