/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 06/05/2003                                                 */
/*                                                                           */
/* File: "object_main.c"                                                     */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "acticast_include.h"

#define BACKGROUND_VALUE 225

#define MY_MIN_BLOB_SIZE 20
#define MY_MAX_NUMBER_OF_BLOBS 100
#define MY_MAX_BLOB_SIZE 30000

/*****************************************************************************/

typedef struct gaussian_struct {
  double *mean;
  double **sigma;
  double **inv;
} GAUSSIAN3;


int interactive_mode(PPM *image, int first, char *message);
int handle_event(XEvent *e);
int allocate_gaussian_color_model(GAUSSIAN3 *model);
int create_gaussian_color_model_from_ppm(PPM *image, GAUSSIAN3 *model,
					   int ignore_value);
double compute_gaussian_likelihood_pixel(char *rgb, GAUSSIAN3 *model);
int gaussian_likelihood_map(double **mat, PPM *image, GAUSSIAN3 *model);
int gaussian_threshold(PPM *t, PPM *image, GAUSSIAN3 *model,
		       double thresh);
int  find_most_likely_blob(BLOB2 *b, int n_blobs, GAUSSIAN3 *model);

/******************************* Globals *************************************/

RAW ImageRaw, TmpRaw;
PPM Image, Data, Density, Threshold, Median, Detected;
BLOB2 Blobs[MY_MAX_NUMBER_OF_BLOBS];

/*****************************************************************************/

int main(int argc, char **argv)
{
  int i;
  char *im_file = NULL, *data_file = NULL;
  int ignore_value = BACKGROUND_VALUE;
  int filter_size = 3;
  int n_blobs, most_likely_blob;
  GAUSSIAN3 model;
  double **mat, thresh = 10.;
  POINT p;

  arg_parse(argc, argv,
            "", "Usage:",
	    "-image %S", &im_file, "set image file",
	    "-data %S", &data_file, "set data image file",
	    "-threshold [%F]", &thresh, "set distance threshold [%lf]",
	    thresh,
	    "-filterSize [%d]", &filter_size, 
	    "set median filter size [2*%d + 1 = %d]", filter_size, 
	    2*filter_size + 1,
	    NULL);

 if(im_file == NULL || data_file == NULL) {
    fprintf(stderr, "No image file and/or data image file specified.\n");
    exit(-1);
  }

  read_ppm_file2(data_file, &Data);
  interactive_mode(&Data, 1, "Data image");
  read_ppm_file2(im_file, &Image);
  ppm2raw(&Image, &ImageRaw, 0);
  interactive_mode(&Image, 0, "Test image");
  allocate_gaussian_color_model(&model);
  create_gaussian_color_model_from_ppm(&Data, &model, ignore_value);
  mat = allocate_matrix(mat, Image.ydim, Image.xdim);
  gaussian_likelihood_map(mat, &Image, &model);
  Density.xdim = Image.xdim;
  Density.ydim = Image.ydim;
  matrix_density_plot(&Density, mat, Image.ydim, Image.xdim, 1, 
		      COLOR_BLACK);
  interactive_mode(&Density, 0, "Gaussian distance map");
  gaussian_threshold(&Threshold, &Image, &model, thresh);
  interactive_mode(&Threshold, 0, "Thresholded image");
  median_filter(&Threshold, &Median, filter_size, 0);
  interactive_mode(&Median, 0, "Median-filtered threshold image");
  for(i = 0; i < MY_MAX_NUMBER_OF_BLOBS; i++) {
    allocate_blob2(&Blobs[i], MY_MAX_BLOB_SIZE);
  }
  n_blobs = new_detect_blobs_ppm(&ImageRaw, &Median, &Blobs[0], 
				 MY_MIN_BLOB_SIZE);
  copy_ppm_image(&Image, &Detected);
  for(i = 0; i < n_blobs; i++) {
    p.x = (int) (Blobs[i].ell.mean[0] + 0.5);
    p.y = (int) (Blobs[i].ell.mean[1] + 0.5);
    draw_ellipse(&Detected, &Blobs[i].ell, COLOR_GREEN);
    draw_big_point(&Detected, &p, COLOR_YELLOW);
  }
  interactive_mode(&Detected, 0, "Detected ball image");

  most_likely_blob = find_most_likely_blob(&Blobs[0], n_blobs, &model);
  p.x = (int) (Blobs[most_likely_blob].ell.mean[0] + 0.5);
  p.y = (int) (Blobs[most_likely_blob].ell.mean[1] + 0.5);
  draw_ellipse(&Image, &Blobs[most_likely_blob].ell, COLOR_CYAN);
  draw_big_point(&Image, &p, COLOR_YELLOW);
  interactive_mode(&Image, 0, "Most likely blob detected");

  exit(1);
}

/*****************************************************************************/

int interactive_mode(PPM *image, int first, char *message)
{
  int done = 0;
  XEvent event;

  if(first) {
    xdisplay_X_DIM = image->xdim;
    xdisplay_Y_DIM = image->ydim;
    ppm2raw(image, &TmpRaw, 0);
    open_xdisplay(TmpRaw.p);
  } else {
    ppm2raw(image, &TmpRaw, 1);
    full_update_xdisplay(TmpRaw.p);
  }

  fprintf(stderr, 
	  "%s. Left button ==> forward\n", message);

  XSelectInput(xdisplay_Display, xdisplay_Window, ButtonPress);

  while(!done) {
    XWindowEvent(xdisplay_Display, xdisplay_Window, ButtonPress, 
		 &event);
    done = handle_event(&event);
  }

  return(done);
}

/*****************************************************************************/

int handle_event(XEvent *e)
{
  int x, y, button;

  x = e->xbutton.x;
  y = e->xbutton.y;
  button = e->xbutton.button;

  if(button == Button1) {
    return(1);
  } else if (button == Button3) {
    return(-1);
  }

  return(0);
}

/*****************************************************************************/

int allocate_gaussian_color_model(GAUSSIAN3 *model)
{
  model->mean = allocate_vector(model->mean, 3);
  zero_vector(model->mean, 3);
  model->sigma = allocate_matrix(model->sigma, 3, 3);
  zero_matrix(model->sigma, 3, 3);
  model->inv = allocate_matrix(model->inv, 3, 3);
  zero_matrix(model->inv, 3, 3);

  return(1);
}

/*****************************************************************************/

int create_gaussian_color_model_from_ppm(PPM *image, GAUSSIAN3 *model,
					   int ignore_value)
{
  int i, j, ri, gi, bi, n = 0;
  double r, g, b, nd;

  for(i = 0; i < image->ydim; i++) {
    for(j = 0; j < image->xdim; j++) {
      ri = convert_char(image->p[j][i][RED]);
      gi = convert_char(image->p[j][i][GREEN]);
      bi = convert_char(image->p[j][i][BLUE]);

      if(!(ri == ignore_value && gi == ignore_value && bi == ignore_value)) {
	r = (double) ri;
	g = (double) gi;
	b = (double) bi;

	model->mean[RED] += r;
	model->mean[GREEN] += g;
	model->mean[BLUE] += b;

	model->sigma[RED][RED] += r*r;
	model->sigma[RED][GREEN] += r*g;
	model->sigma[RED][BLUE] += r*b;
	model->sigma[GREEN][GREEN] += g*g;
	model->sigma[GREEN][BLUE] += g*b;
	model->sigma[BLUE][BLUE] += b*b;

	n++;
	//fprintf(stderr, "%d %d %d\n", ri, gi, bi);
      }
    }
  }

  nd = ((double) n);

  model->mean[RED] /= nd;
  model->mean[GREEN] /= nd;
  model->mean[BLUE] /= nd;
  
  r = model->mean[RED];
  g = model->mean[GREEN];
  b = model->mean[BLUE];

  model->sigma[RED][RED] = nd/(nd - 1.)*(model->sigma[RED][RED]/nd - r*r);
  model->sigma[RED][BLUE] = nd/(nd - 1.)*(model->sigma[RED][BLUE]/nd - r*b);
  model->sigma[RED][GREEN] = nd/(nd - 1.)*(model->sigma[RED][GREEN]/nd - r*g);
  model->sigma[GREEN][RED] = model->sigma[RED][GREEN];
  model->sigma[GREEN][GREEN] = 
    nd/(nd - 1.)*(model->sigma[GREEN][GREEN]/nd - g*g);
  model->sigma[GREEN][BLUE] = 
    nd/(nd - 1.)*(model->sigma[GREEN][BLUE]/nd - g*b);
  model->sigma[BLUE][RED] = model->sigma[RED][BLUE];
  model->sigma[BLUE][GREEN] = model->sigma[GREEN][BLUE];
  model->sigma[BLUE][BLUE] = 
    nd/(nd - 1.)*(model->sigma[BLUE][BLUE]/nd - b*b);
  
  inverse(model->inv, model->sigma, 3);
  

  return(1);
}

/*****************************************************************************/

double compute_gaussian_likelihood_pixel(char *rgb, GAUSSIAN3 *model)
{
  double v[3], tmp[3], dist;

  v[RED] = model->mean[RED] - ((double) convert_char(rgb[RED]));
  v[GREEN] = model->mean[GREEN] - ((double) convert_char(rgb[GREEN]));
  v[BLUE] = model->mean[BLUE] - ((double) convert_char(rgb[BLUE]));

  matrix_vector_product(&tmp[0], model->inv, &v[0], 3, 3);
  dist = dot(&v[0], &tmp[0], 3);

  return(dist);
}

/*****************************************************************************/

int gaussian_likelihood_map(double **mat, PPM *image, GAUSSIAN3 *model)
{
  int i, j;

  for(i = 0; i < image->ydim; i++)
    for(j = 0; j < image->xdim; j++) {
      mat[image->ydim-i-1][j] = 
	compute_gaussian_likelihood_pixel(&image->p[j][i][0], model);
    }

  return(1);
}

/*****************************************************************************/

int gaussian_threshold(PPM *t, PPM *image, GAUSSIAN3 *model,
		       double thresh)
{
  int i, j;
  double dist;

  limit_dimensions(image);
  copy_dimensions(image, t);

  for(i = 0; i < image->ydim; i++)
    for(j = 0; j < image->xdim; j++) {
      dist =
	compute_gaussian_likelihood_pixel(&image->p[j][i][0], model);
      if(dist < thresh) {
	t->p[j][i][RED] = (char) 255;
	t->p[j][i][GREEN] = (char) 255;
	t->p[j][i][BLUE] = (char) 255;
      } else {
	t->p[j][i][RED] = (char) 0;
	t->p[j][i][GREEN] = (char) 0;
	t->p[j][i][BLUE] = (char) 0;
      }
    }

  return(1);
}

/*****************************************************************************/

int  find_most_likely_blob(BLOB2 *b, int n_blobs, GAUSSIAN3 *model)
{
  int i, j, blob_id = 0;
  char rgb[3];
  double dist, min_dist = 1000000000.;

  for(i = 0; i < n_blobs; i++) {
    dist = 0.;
    for(j = 0; j < b[i].n; j++) {
      rgb[RED] = (char) ((int) b[i].rgb_t[RED][j]);
      rgb[GREEN] = (char) ((int) b[i].rgb_t[GREEN][j]);
      rgb[BLUE] = (char) ((int) b[i].rgb_t[BLUE][j]);
      dist += compute_gaussian_likelihood_pixel(&rgb[0], model);
    }
    dist /= ((double) b[i].n);

    fprintf(stderr, "%d (%d): Distance = %lf\n", i, b[i].n, dist);

    if(dist < min_dist) {
      min_dist = dist;
      blob_id = i;
    }
  }
   
  return(blob_id);
}

/*****************************************************************************/
